#ifndef WIN32
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#else
#include <winsock2.h>
#endif

#include "Socket.h"
#include "Streamable.h"

Socket::Socket ( int domain, int type, int protocol ) : Initialized ( false )
{
  SocketDescriptor = socket ( domain, type, protocol );

  Initialized = ( SocketDescriptor > 0 );
  Domain = domain;
}

Socket::Socket ( int fd, const string& clientip ) : Initialized ( true ),
                                                    SocketDescriptor ( fd ),
                                                    ClientIP ( clientip )
{

}

Socket::~Socket ()
{

}

void Socket::Connect ( const string& hostname, u_short port )
{
  if ( !Initialized )
  {
    throw ( 1 );
  }

  struct sockaddr_in servaddr;

  memset ( &servaddr, 0, sizeof ( servaddr ) );
  servaddr.sin_family = Domain;
  servaddr.sin_port = htons ( port );
  if ( isdigit ( hostname [ 0 ] ) )
  {
    servaddr.sin_addr.s_addr = inet_addr ( hostname.c_str() );
  }    
  else
  {
    struct hostent* hostStruct;
    struct in_addr* hostNode;

    hostStruct = gethostbyname ( hostname.c_str() );

    if ( hostStruct )
    {
      hostNode = ( struct in_addr* ) hostStruct -> h_addr;
      servaddr.sin_addr.s_addr = hostNode -> s_addr;
    }
    else
    {
      throw ( 1 );
    }
  }

  if ( connect ( SocketDescriptor, ( struct sockaddr* ) &servaddr,
                 sizeof ( servaddr ) ) < 0 )
  {
    throw ( 1 );
  }
}

void Socket::Bind ( u_short port )
{
  if ( !Initialized )
  {
    throw ( 1 );
  }

  struct sockaddr_in servaddr;

  memset ( &servaddr, 0, sizeof ( servaddr ) );
  servaddr.sin_family = Domain;
  servaddr.sin_addr.s_addr = htonl ( INADDR_ANY );
  servaddr.sin_port = htons ( port );

  if ( bind ( SocketDescriptor, ( struct sockaddr* ) &servaddr,
              sizeof ( servaddr ) ) == -1 )
  {
    throw ( 1 );
  }
}

void Socket::Listen ( int queue )
{
  if ( !Initialized )
  {
    throw ( 1 );
  }

  listen ( SocketDescriptor, queue );
}

Socket Socket::Accept ()
{
  if ( !Initialized )
  {
    throw ( 1 );
  }

  struct sockaddr_in clientaddr;

#ifndef WIN32
  unsigned int clientlen;
#else
  int clientlen;
#endif;

  clientlen = sizeof ( clientaddr );

  int fd;
  fd = accept ( SocketDescriptor, ( struct sockaddr* ) &clientaddr,
                &clientlen );

  string clientip;

  clientip = inet_ntoa ( * ( struct in_addr * )  &clientaddr.sin_addr.s_addr );
  Socket socket ( fd, clientip );

  return ( socket );
}

void Socket::Close()
{
#ifndef WIN32
  close ( SocketDescriptor );
#else
  closesocket ( SocketDescriptor );
#endif
}

void Socket::Read ( void* data, int size )
{
  int readleft, readdone;
  char* message;

  message = ( char * ) data;
  readleft = size;

  fd_set rset;

  while ( readleft > 0 )
  {
    FD_ZERO ( &rset );
    FD_SET ( SocketDescriptor, &rset );

    if ( select ( SocketDescriptor + 1, &rset, NULL, NULL, NULL ) <= 0 )
    {
      throw ( 1 );
    }

    readdone = recv ( SocketDescriptor, message, readleft, 0 );

    if ( readdone == 0 )
    {
      break;
    }

    readleft -= readdone;
    message += readdone;
  }
}

void Socket::Write ( const void* data, int size )
{
  int nleft;
  int nwritten;
  char* ptr;

  ptr = ( char * ) data;
  nleft = size;

  while ( nleft > 0 )
  {
    nwritten = send ( SocketDescriptor, ptr, nleft, 0 );

    nleft -= nwritten;
    ptr += nwritten;
  }
}

Socket& Socket::operator << ( char data )
{
  Write ( ( void* ) &data, sizeof ( data ) );
  return ( *this );
}

Socket& Socket::operator << ( int data )
{
  unsigned long int convert = htonl ( data );

  Write ( ( void* ) &convert, sizeof ( convert ) );
  return ( *this );
}

Socket& Socket::operator << ( long data )
{
  unsigned long int convert = htonl ( data );

  Write ( ( void* ) &convert, sizeof ( convert ) );
  return ( *this );
}

Socket& Socket::operator << ( float data )
{
  Write ( ( void* ) &data, sizeof ( data ) );
  return ( *this );
}

Socket& Socket::operator << ( double data )
{
  Write ( ( void* ) &data, sizeof ( data ) );
  return ( *this );
}

Socket& Socket::operator>> ( char& data )
{
  Read ( ( void* ) &data, sizeof ( data ) );
  return ( *this );
}

Socket& Socket::operator>> ( int& data )
{
  unsigned long int convert;

  Read ( ( void* ) &convert, sizeof ( convert ) );
  data = ntohl ( convert );
  return ( *this );
}

Socket& Socket::operator>> ( long& data )
{
  unsigned long int convert;

  Read ( ( void* ) &convert, sizeof ( convert ) );
  data = ntohl ( convert );
  return ( *this );
}

Socket& Socket::operator>> ( float& data )
{
  Read ( ( void* ) &data, sizeof ( data ) );
  return ( *this );
}

Socket& Socket::operator>> ( double& data )
{
  Read ( ( void* ) &data, sizeof ( data ) );
  return ( *this );
}

Socket& Socket::operator<< ( const Streamable& streamable )
{
  streamable.Marshall ( *this );
  return ( *this );
}

Socket& Socket::operator>> ( Streamable& streamable )
{
  streamable.UnMarshall ( *this );
  return ( *this );
}


